/*
Decomposition Tree implementation, version 3.1.0 DT_DEBUG VERSION
Zvika Brakerski, Jon Feldman, Guy Even
Summer, 2000
*/

/*
This is an implementation of constructing a decomposition tree given an hmetis format file.
The construction is done using repeated bisections using the hmetis component.
*/

// Library include file
//#include <?>


#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <time.h>

// Orientation tree related definitions
#include "ot.h"
#include "hyperorient.h"

int SEPARATE_FLAG;

//void HMETIS_PartRecursive(int, int, int *, int *, int *, int *, int, int, int *, int *, int *);

// Definitions
#define NPARTS 2
//#define DT_DEBUG
//#define DT_DEBUG2
#define HPARAM

#define NRUNS 10
#define CTYPE 4
#define RTYPE 1
#define VCYCLE 0
#define RECONST 0


#define LINE 512

typedef struct linkedlist { 
// This structures holds an element of the linked list used to store hyperedges tomporarily
  struct linkedlist *next;
  int data;
} element;


char *fgets_no_comments(char *s, int size, FILE *stream) 
// This function wraps fgets, this is just fgets that ignores lines that start with %(remarks)
{
  char *temp=fgets(s, size, stream);
  if (temp && temp[0]=='%')
    return fgets_no_comments(s, size, stream);
  return temp;
}

void extract_from_file(FILE *hg, int *nvtxs, int *nhedges, int **eptr, int **eind) 
// This function extracts the graph properties from an hmetis type file
// Note that this only works for UNWEIGHTED hypergraphs
// Also note that on the file, the nodes are numbered 1,...,nhedges wheres in the procedure representation they are indexed 0,...,nhedges-1
{
  char line[LINE+1], *s;
  element **a, *temp, *temp2;
  int i, num;
  int eind_size=0;

  fgets_no_comments(line, LINE, hg); // Getting the next line
  sscanf(line, "%d %d", nhedges, nvtxs); // Getting the number of hyperedges and vertexes
  *eptr=malloc((*nhedges+1)*sizeof(int)); // Allocating room for eptr
  a = calloc(*nhedges, sizeof(element *)); // Temporary array to store hyperedges in linked list form
  for (i=0; i<*nhedges; i++) { // For each hyperedge

#ifdef DEBUG
    printf("%i\n",i);
#endif

    (*eptr)[i]=eind_size; // The starting line for the i-th hyperedge
    s = line;
    fgets_no_comments(line, LINE, hg); // Getting the next hyperedge(line)

    while (sscanf(s, "%d", &num)==1) { // Extracting the vertexes of the hyperedge into a linked list
      while (!isdigit(*s)) s++;
      eind_size++;
      temp = malloc(sizeof(element));
      temp->next = a[i];
      temp->data = --num;
      a[i]=temp;
      while (isdigit(*s)) s++;
    }
  }
  
  (*eptr)[i]=eind_size; // Last index in eptr points over the end of eind
  *eind = malloc(eind_size*sizeof(int)); // Allocating room for eind
  num=0; // Auxiliary variable
  for(i=0;i<*nhedges;i++)
    for(temp=a[i]; temp; temp=temp->next, free(temp2))
      (*eind)[num++]=(temp2=temp)->data;

  free(a);

  return;
      
}

dt_node *do_decomposition_SEP(int *node_ids, // An array that holds the original names of the vertexes
			  int nvtxs, // Number of vertexes
			  int nhedges, // Number of hyperedges
			  int *eptr, // A pointer array(hmetis format)
			  int *eind, // An index array(hmetis format)
			  int ubfactor, // relative imbalance factor(hmetis format)
			  hypergraph *g // The original hypergraph
			  )
{
        dt_node dtn, *pdtn; // A decomposition tree node to which we make the changes and a pointer to be used as a returned value
	
#ifdef HPARAM
	int temp[9] = {1, NRUNS, CTYPE, RTYPE, VCYCLE, RECONST, 0, -1, 0};
	srand(time(NULL));
#else
	int temp[1]={0};
#endif

	//	printf("Doing NEW scheme\n");

	dtn.lca_flag = -1;
	dtn.inner_cut_temp = 0;
	dtn.orientation = 0;
	dtn.orientation_worst = 0;
	
	if (nvtxs==0) // If there are no vertexes(an empty decomposition tree)
		return NULL;

	if (nvtxs==1) // Only one node(leaf)
	{
#ifdef DT_DEBUG
		printf("node %d added\n", node_ids[0]);
#endif
		// Setting node properties
		dtn.child[0]=dtn.child[1]=NULL;
		dtn.v = node_ids[0];
		dtn.inner_cut = 0;
		dtn.subtree_size = 1;
		
		pdtn = malloc(sizeof(dt_node));
		*pdtn=dtn;
		g->ref_array[node_ids[0]]=pdtn; // Updating the ref_array
		return pdtn;
	}

	{ 
	  // If there's more than one node

	        int i, j; // Auxiliary variables
	        int num1, num2; // Auxiliary counters
		int *results=malloc(nvtxs*sizeof(int)); // An array to store partition results(0/1)
		int nvtxs1=0, nvtxs2=0, nhedges1=0, nhedges2=0; // Number of vertexes and hyperedges in each subgraph
		int *node_ids1 = malloc(nvtxs*sizeof(int)), *node_ids2 = malloc(nvtxs*sizeof(int)); // Allocating node_ids arrays for subgraphs
		int *eptr1=malloc((nhedges+1)*sizeof(int)), *eind1=malloc(eptr[nhedges]*sizeof(int)); // Allocating enough room for eind arrays for both subgraphs
		int *eptr2=malloc((nhedges+1)*sizeof(int)), *eind2=malloc(eptr[nhedges]*sizeof(int)); // Allocating enough room for eptr arrays for both subgraphs 
		int *convert=malloc(nvtxs*sizeof(int)); // This is an array used to convert the indexes of the original nodes into indexes of nodes in the subgraphs(as in all cases the node indexes should be 0,...,nhedges-1)

		HMETIS_PartRecursive(nvtxs, nhedges, NULL, eptr, eind, NULL, NPARTS, ubfactor, temp, results, &(dtn.inner_cut)); // Partitioning using hmetis
		dtn.v = NOT_LEAF;
		dtn.subtree_size = nvtxs;

		for(i=0; i<nvtxs; i++) // Creating convert and node_id arrays
			if (results[i]==0)
			{
				convert[i] = nvtxs1;
				node_ids1[nvtxs1++]=node_ids[i];
			}
			else
			{
				convert[i] = nvtxs2;
				node_ids2[nvtxs2++]=node_ids[i];
			}
				
		if (!(nvtxs1 && nvtxs2)) // In case we got a bad partition(one part has 0 nodes) we partition arbiterally
		{
#ifdef DT_DEBUG2
			
		  // Printing an error message

		        fprintf(stderr, "hmetis fault:\n");
			fprintf(stderr, "{ ");

			for(i=0; i<nvtxs1; i++)
				fprintf(stderr, "%d ", node_ids1[i]);

			fprintf(stderr, "}\n");
			fprintf(stderr, "{ ");

			for(i=0; i<nvtxs2; i++)
				fprintf(stderr, "%d ", node_ids2[i]);
			fprintf(stderr, "}\n\n");
#endif
			
			// Repartitioning arbiterally(odd/even)
			nvtxs1 = nvtxs2 = 0;
			
			for(i=0; i<nvtxs; i++)
			  if ((i%2)==0) 
			{
				convert[i] = nvtxs1;
				node_ids1[nvtxs1++]=node_ids[i];
			}
			else
			{
				convert[i] = nvtxs2;
				node_ids2[nvtxs2++]=node_ids[i];
			}			    
				
		}
		
		
#ifdef DT_DEBUG
		// Outputting results of partition
		printf("partitioning:\n");
		printf("{ ");
		for(i=0; i<nvtxs1; i++)
			printf("%d ", node_ids1[i]);
		printf("}\n");
		printf("{ ");
		for(i=0; i<nvtxs2; i++)
			printf("%d ", node_ids2[i]);
		printf("}\n\n");
#endif DT_DEBUG

		// Creating eptr and eind arrays

		eptr1[0]=eptr2[0]=0; // Initializing the first index
		for(i=0; i<nhedges; i++) // For each hyperedge in the original hypergraph
		{
		  num1=eptr1[nhedges1];
		  num2=eptr2[nhedges2];
		  for (j=eptr[i]; j<eptr[i+1]; j++) { // For each noder associated with this hyperedge
		    if (results[eind[j]]==0) // If it belongs to the first part
		      eind1[num1++]=convert[eind[j]];
		    else // If it belongs to the second part
		      eind2[num2++]=convert[eind[j]];
		  }
		    
		  if ((num1-eptr1[nhedges1]) > 1) // If we have an actual hyperedge in the first subgraph(meaning one with more than 1 node)
		      eptr1[++nhedges1]=num1;

		    if ((num2-eptr2[nhedges2]) > 1) // If we have an actual hyperedge in the second subgraph(meaning one with more than 1 node)
		      eptr2[++nhedges2]=num2;

		}
		
		free(results);
	
		// Calling the function recursively for children
		dtn.child[0] = do_decomposition_SEP(node_ids1, nvtxs1, nhedges1, eptr1, eind1, ubfactor, g);
		dtn.child[1] = do_decomposition_SEP(node_ids2, nvtxs2, nhedges2, eptr2, eind2, ubfactor, g);
		
		// Freeing memory
		free(node_ids1);
		free(node_ids2);
		free(eptr1);
		free(eptr2);
		free(eind1);
		free(eind2);

	}


	dtn.inner_cut = 0;


	pdtn = malloc(sizeof(dt_node));
	*pdtn=dtn;
	
	// Assigning value for parent field of children
	for((*temp)=0; (*temp)<2; (*temp)++)
		if (pdtn->child[*temp])
			(pdtn->child[*temp])->parent = pdtn;

	return pdtn;
}


dt_node *create_decomposition_tree(FILE *hg, // the given hyper-graph
				int ubfactor, // relative imbalance factor
				hypergraph *g // Hypergraph structure  
				)
{
// Notes on arguments:
// ===================
// hg-
//	This is a file pointer to the hyper-graph for which we want to create a decomposition tree.
//	
//
// ubfactor-
//	Properties of ubfactor are characterised in the hmetis documentation.
//	
//
//
// Returned value:
// ===============
// Root of the decomposition tree.

	int i; // Auxiliary variables
	int nvtxs, nhedges; // Number or vertexes and hyperedges
	int *eptr, *eind; // Hyperedges
	int *node_ids = malloc(nvtxs*sizeof(int)); // An array that holds the node numbers
	dt_node *root; // Tree root

	extract_from_file(hg, &nvtxs, &nhedges, &eptr, &eind);
 
	node_ids = malloc(nvtxs*sizeof(int));

	g->ref_array = malloc(nvtxs*sizeof(int));
	g->size = nvtxs;
	g->hyperedges=nhedges;
	g->hyperedge_array=eptr;
	g->vertex_array=eind;

	for(i=0; i<nvtxs; i++)
		node_ids[i]=i;

	if(SEPARATE_FLAG)
	  root = do_decomposition_SEP(node_ids, nvtxs, nhedges, eptr, eind, ubfactor, g);
	else
	  root = do_decomposition(node_ids, nvtxs, nhedges, eptr, eind, ubfactor, g);


	root->parent = NULL;
	
	free(node_ids);


	return root;
}
	



dt_node *do_decomposition(int *node_ids, int nvtxs, int nhedges, int *eptr, int *eind, int ubfactor, hypergraph *g)
{


	dt_node dtn, *pdtn;
	
#ifdef HPARAM
	int temp[9] = {1, NRUNS, CTYPE, RTYPE, VCYCLE, RECONST, 0, -1, 0};
	srand(time(NULL));
#else
	int temp[1]={0};
#endif

	dtn.lca_flag = -1;
	dtn.inner_cut_temp = 0;
	dtn.orientation = 0;
	dtn.orientation_worst = 0;

	//  printf("Doing OLD scheme\n");
	
	if (nvtxs==0)
		return NULL;

	if (nvtxs==1)
	{
#ifdef DT_DEBUG
		printf("node %d added\n", node_ids[0]);
#endif
		
		dtn.child[0]=dtn.child[1]=NULL;
		dtn.v = node_ids[0];
		dtn.inner_cut = 0;
		dtn.subtree_size = 1;
		
		pdtn = malloc(sizeof(dt_node));
		*pdtn=dtn;
		g->ref_array[node_ids[0]]=pdtn;
		return pdtn;
	}

	{
		int i, j;
		int b=1;
		int *results=malloc(nvtxs*sizeof(int));
		int nvtxs1=0, nvtxs2=0, nhedges1=0, nhedges2=0;
		int *node_ids1 = malloc(nvtxs*sizeof(int)), *node_ids2 = malloc(nvtxs*sizeof(int));
		int *eptr1=malloc((nhedges+1)*sizeof(int)), *eind1=malloc(eptr[nhedges]*sizeof(int));
		int *eptr2=malloc((nhedges+1)*sizeof(int)), *eind2=malloc(eptr[nhedges]*sizeof(int));
		int *convert=malloc(nvtxs*sizeof(int));

		HMETIS_PartRecursive(nvtxs, nhedges, NULL, eptr, eind, NULL, NPARTS, ubfactor, temp, results, &(dtn.inner_cut));
		dtn.v = NOT_LEAF;
		dtn.subtree_size = nvtxs;

		for(i=0; i<nvtxs; i++)
			if (results[i]==0)
			{
				convert[i] = nvtxs1;
				node_ids1[nvtxs1++]=node_ids[i];
			}
			else
			{
				convert[i] = nvtxs2;
				node_ids2[nvtxs2++]=node_ids[i];
			}
				
		if (!(nvtxs1 && nvtxs2))
		{
#ifdef DT_DEBUG2
			fprintf(stderr, "hmetis fault:\n");
			fprintf(stderr, "{ ");

			for(i=0; i<nvtxs1; i++)
				fprintf(stderr, "%d ", node_ids1[i]);

			fprintf(stderr, "}\n");
			fprintf(stderr, "{ ");

			for(i=0; i<nvtxs2; i++)
				fprintf(stderr, "%d ", node_ids2[i]);
			fprintf(stderr, "}\n\n");
#endif
			
			nvtxs1 = nvtxs2 = 0;
			
			for(i=0; i<nvtxs; i++)
			  if ((i%2)==0) 
			{
				convert[i] = nvtxs1;
				node_ids1[nvtxs1++]=node_ids[i];
			}
			else
			{
				convert[i] = nvtxs2;
				node_ids2[nvtxs2++]=node_ids[i];
			}			    
				
		}
		
		
#ifdef DT_DEBUG
		printf("partitioning:\n");
		printf("{ ");
		for(i=0; i<nvtxs1; i++)
			printf("%d ", node_ids1[i]);
		printf("}\n");
		printf("{ ");
		for(i=0; i<nvtxs2; i++)
			printf("%d ", node_ids2[i]);
		printf("}\n\n");
#endif DT_DEBUG

		eptr1[0]=eptr2[0]=0;
		for(i=0; i<nhedges; i++)
		{
		  b=1;
			for (j=eptr[i]; j<eptr[i+1]; j++)
				if (results[eind[j]]!=results[eind[eptr[i]]]) b=0;
			if (b) {
				if (results[eind[eptr[i]]] == 0)
				{
					for (j=eptr[i]; j<eptr[i+1]; j++)
						eind1[eptr1[nhedges1]+j-eptr[i]]=convert[eind[j]];
					nhedges1++;
					eptr1[nhedges1]=eptr1[nhedges1-1]+eptr[i+1]-eptr[i];
				}
				else
				{
					for (j=eptr[i]; j<eptr[i+1]; j++)
						eind2[eptr2[nhedges2]+j-eptr[i]]=convert[eind[j]];
					nhedges2++;
					eptr2[nhedges2]=eptr2[nhedges2-1]+eptr[i+1]-eptr[i];
				}
			}
		}
		
		free(results);
	
		dtn.child[0] = do_decomposition(node_ids1, nvtxs1, nhedges1, eptr1, eind1, ubfactor, g);
		dtn.child[1] = do_decomposition(node_ids2, nvtxs2, nhedges2, eptr2, eind2, ubfactor, g);
		
		free(node_ids1);
		free(node_ids2);
		free(eptr1);
		free(eptr2);
		free(eind1);
		free(eind2);

	}


	dtn.inner_cut = 0;


	pdtn = malloc(sizeof(dt_node));
	*pdtn=dtn;
	
	for((*temp)=0; (*temp)<2; (*temp)++)
		if (pdtn->child[*temp])
			(pdtn->child[*temp])->parent = pdtn;

	return pdtn;
}

