#include <time.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "ot.h"
#include "hyperorient.h"



#define LEFT_CUT 0
#define RIGHT_CUT 1
#define TIMING_INTERVAL 100000

or_tree * hyp_OrientRecursive(//INPUTS:
			      dt_node * dt_root,           //root of dt
			      dt_node * dt_proto,          //node in dt corresponding to this node of or tree
			      hyp_LCAinfo ** lca_info,
			      hypergraph * g, inverse_hypergraph * h,
 			      //OUTPUTS:
			      or_tree ** solution,
			      int * local_cost,     
			      int * left_cut,
			      int * right_cut
			      )
{
  int child;
  int winner;
  int cost[2];
  int possible_left_cut[2][2];
  int possible_right_cut[2][2];
  int possible_local_cost[2][2];
  int b;
  int aaa,bbb,ccc;

  //Holds the two possible orientation trees
  or_tree * possible_solution[2];
  
  //If this is a leaf of the dt, then we need to do something completely different:
  if(dt_proto->v != NOT_LEAF)
    {
      //The orientation is irrelevant, so just return NULL
      (*solution = NULL);

      (*left_cut) = 0;
      (*right_cut) = 0;

      //Compute the outer cuts
      hyp_OuterCutsOfSingleVertex(//INPUTS:
				  dt_proto->v, g, h, lca_info,
#ifdef OUTER_CUT_INFO
				  dt_root,
#endif

				  //OUTPUTS:
				  left_cut, right_cut);
#ifdef OVERFLOW_TEST
      printf("Left Cut: %i, Right Cut: %i, Number of edges: %i\n",
	     *left_cut,*right_cut,NumberOfHyperedgesCoveringVertex(dt_proto->v,h)); 
#endif


      (*local_cost) = (*left_cut);

      return (*solution);

    }

  //allocate memory for orientation trees
  possible_solution[0] = New(or_tree);
  possible_solution[1] = New(or_tree);

  //for each possible orientation b
  for (b=1; b>=0; b--)
    {
      //orientation of root doesn't matter, so set it to 1, and skip 0
      //(Cuts running time in half)
      if(dt_proto == dt_root) b=0;


      //Set this orientation into dt for children's knowledge
      dt_proto->orientation = b;
      
      //   recursively find best orientation, local cost, and outer cuts
      //   of left and right children, given orientation b
      for(child=0; child<=1; child++)
	{
	  hyp_OrientRecursive(//INPUTS
			      dt_root, 
			      dt_proto->child[child],
			      lca_info,
			      g,h,
			      //OUTPUTS
			      &(possible_solution[b]->child[child]),
			      &(possible_local_cost[b][child]),
			      &(possible_left_cut[b][child]),
			      &(possible_right_cut[b][child])
#ifdef PROGRESS_METER
			      ,ot_size
#endif
			      );
 
#ifdef PROGRESS_METER
	  ++path[depth];
	  //	  path[depth] = (path[depth] - '0') % 4 + '0';
#endif
	  
	}
        
      //   Compute local cost of this orientation based on values returned
      //   by two children.
      
      //The local cost depends on the problem being solved (Cut-width or Linear Arrangement)
      cost[b] = LocalCostOnTheFly( possible_local_cost[b][b],
				       possible_local_cost[b][1-b],
				       possible_right_cut[b][b],
				       possible_left_cut[b][1-b],
				       dt_proto->inner_cut,
				       dt_proto->child[b]->subtree_size,
				       dt_proto->child[1-b]->subtree_size);
    }
  
  //determine winner, set orientation;
  if(dt_proto == dt_root)
    winner = 0;
  else
    winner = (cost[0] >= cost[1]);
  
#ifdef OVERFLOW_TEST

  if(cost[1-winner] > 53870912) printf("OVERFLOW!!!\n");

#endif

#ifdef COST_TEST

  if((cost[0] <=0) || (cost[1] <= 0)) printf("AHA!!!!\n");


#endif

  possible_solution[winner]->orientation = winner;

  //Take better one as return value of local cost.
  (*local_cost) = cost[winner];
  
  //compute left and right cuts, return them.
  (*left_cut) = - dt_proto->inner_cut;
  (*left_cut) += possible_left_cut[winner][0];
  (*left_cut) += possible_left_cut[winner][1];
  
  //(*left_cut) = 
  //  possible_left_cut[winner][0] 
  //  + possible_left_cut[winner][1] 
  //  - dt_proto->inner_cut;

  //compute right and right cuts, return them.
  (*right_cut) = - dt_proto->inner_cut;
  (*right_cut) += possible_right_cut[winner][0];
  (*right_cut) += possible_right_cut[winner][1];


  //(*right_cut) = 
  //  possible_right_cut[winner][0] 
  //  + possible_right_cut[winner][1] 
  //  - dt_proto->inner_cut;

  //reset this orientation for posterity
  dt_proto->orientation = -1;
  
  //Delete the or_tree of the loser
  if(dt_proto != dt_root)
    FreeOrTree(possible_solution[1-winner]);

  //return solution of winner
  (*solution) = possible_solution[winner];
  
  
#ifdef PROGRESS_METER
  path[depth] = NULL;
  depth -= 1;
#endif

  //return local cost as ret value
  return (*solution);

}

//Frees an or_tree
void FreeOrTree(or_tree * t)
{
  if(t != NULL)
    {
      FreeOrTree(t->child[0]);
      FreeOrTree(t->child[1]);
	
      free(t);
    }
}


int LocalCostOnTheFly(int local_cost_lefttree,
		      int local_cost_righttree,
		      int right_cut_lefttree,
		      int left_cut_righttree,
		      int inner_cut,
		      int size_lefttree,
		      int size_righttree )
{

#ifdef LINEAR_ARRANGEMENT

  //Linear Arrangement:
  
  int a = local_cost_lefttree + local_cost_righttree;
  int b = ( (right_cut_lefttree - inner_cut) * size_righttree );
  int c = ( (left_cut_righttree - inner_cut) * size_lefttree );


#ifdef OVERFLOW_TEST

  //printf("%i,%i,%i;",a,b,c);

  if((a<0)||(b<0)||(c<0)) printf("LOCAL COST OVERFLOW!!!\n");

#endif
  
  return a+b+c;

  //  return local_cost_lefttree + local_cost_righttree 
  //  + ( (right_cut_lefttree - inner_cut) * size_righttree )
  //  + ( (left_cut_righttree - inner_cut) * size_lefttree );
  

#endif


}





void OrientDtFromOrTree(dt_node * dt, or_tree * t)
{

  if(t != NULL)
    {
      
      dt->orientation = t->orientation;

      OrientDtFromOrTree(dt->child[0], t->child[0]);
      OrientDtFromOrTree(dt->child[1], t->child[1]);
      
      free(t);
    }
}



or_tree * hyp_OrientOnTheFly(hypergraph * g, dt_node * dt)
{

  hyp_LCAinfo ** lca_info = hyp_PreProcessLCA(g, dt);
  inverse_hypergraph * h = InverseHypergraph(g);

#ifdef PROGRESS_METER
  int ot_size = SizeOfOt(dt, 0);
#endif

  or_tree * ret;
  or_tree ** solution = New(or_tree*);
  int * local_cost = New(int); 
  int * left_cut = New(int);
  int * right_cut = New(int);
  
  
  ret = hyp_OrientRecursive(//INPUTS:
			    dt,           //root of dt
			    dt,          //node in dt corresponding to this node of ot
			    lca_info,
			    g,h,
			    //OUTPUTS:
			    solution,
			    local_cost,     
			    left_cut,
			    right_cut
#ifdef PROGRESS_METER
			    //info for progress:
			    ,ot_size
#endif
			    );
  

#ifdef LOCAL_COST_AT_ROOT
  printf("Local Cost At Root: %i,",*local_cost);
#endif

  free(local_cost);
  free(left_cut);
  free(right_cut);

  FreeInverseHypergraph(h);
  
  return ret;

}











int hyp_OrientAndEvaluateOnTheFly(hypergraph * g, dt_node * dt)
{

  OrientDtFromOrTree(dt, hyp_OrientOnTheFly(g, dt));
  
  return hyp_CostOfOrientation(dt, g);
  
}








//Need the following precompiler macros:
//DtNodeOfVertex(vertex v)
//SizeOfHyperedge(hyperedge e)
//NameOfVertex(hyperedge e, int i)
//NameOfHyperedge(vertex v, int i)
//WeightOfHyperedge(hyperedge e)
//NumberOfHyperedgesInHypergraph(hypergraph * g)
//NumberOfNodesInHypergraph(hypergraph * g)


hyp_LCAinfo ** hyp_PreProcessLCA(hypergraph * g, dt_node * dt)
{
  int i;
  int size;
  hyp_LCAinfo ** ret;
  
  ret = NewArray(hyp_LCAinfo *, g->hyperedges);
  
  for(i=0; i<g->hyperedges; i++)
    ret[i] = hyp_PreProcessLCAForSingleHyperedge(i, g, dt);

  //cleas the lca info from the dt that was added by hyp_PreProcessLCAForSingleHyperedge
  ClearLocalCosts(dt);

#ifdef DEBUG
  VerifyInnerCuts(g,dt);
#endif

  return ret;

}








//preprocessing for the lca info of hyperedges; performs the
//calculations of inner cuts with respect to this hyperedge and makes
//calculations of outer cuts more efficient.  NOTE: Permanently alters
//lca_flag; must be reset after this process.
hyp_LCAinfo * hyp_PreProcessLCAForSingleHyperedge(hyperedge e, hypergraph * g, dt_node * dt)
{

  dt_node * dt_temp;
  int flag;
  int i;
  hyp_LCAinfo * ret;
  
  //Mark the lca_flag field with the name of the hyperedge for each
  //node in the dt that is an ancestor of a hyperedge vetrtex.
  for(i=0; i < SizeOfHyperedge(e,g); i++)
    for( dt_temp = g->ref_array[g->vertex_array[g->hyperedge_array[e] + i]];
	 dt_temp != NULL;
	 dt_temp = dt_temp->parent )
      dt_temp->lca_flag = (int) e;
  
  //recursively build the LCA structure
  ret = hyp_PreProcessLCAForSingleVertexRecursive(dt, e);

  //set the inner cut of the root of the structure
  ret->corr_dt->inner_cut += 1;   //UNWEIGHTED


#ifdef OVERFLOW_TEST

  if(ret->corr_dt->inner_cut < 0) printf("OVERFLOW ON INNER CUT!!!");

#endif


  return ret;
}

//recursive subprocedure of hyp_PreProcessLCAForSingleHyperedge
hyp_LCAinfo * hyp_PreProcessLCAForSingleVertexRecursive(dt_node * dt, hyperedge e)
{
  hyp_LCAinfo * ret = New(hyp_LCAinfo);
  dt_node * dt_temp;

  dt_temp = FindNextBranchingPoint(dt, e);

#ifdef DEBUG
  if(dt_temp != dt) printf("Branching Point different\n");

#endif

  ret->corr_dt = dt_temp;
  
  if(dt_temp->v != NOT_LEAF)
    {
      ret->child[0] = NULL;
      ret->child[1] = NULL;
    }
  else
    {
      ret->child[0] = hyp_PreProcessLCAForSingleVertexRecursive(dt_temp->child[0], e);
      ret->child[1] = hyp_PreProcessLCAForSingleVertexRecursive(dt_temp->child[1], e);
    }

  return ret;

}



//Finds the next branching point in the dt with respect to this hyperedge
//returns root if root is a branching point.
dt_node * FindNextBranchingPoint(dt_node * root, hyperedge e)
{
  if(root->v != NOT_LEAF)
    return root;

  if(root->child[0]->lca_flag == (int) e)
    if(root->child[1]->lca_flag == (int) e)
      return root;
    else
      return FindNextBranchingPoint(root->child[0], e);
  else
    return FindNextBranchingPoint(root->child[1], e);
  
}

//Finds the outer cuts of a given vertex under the orientation
//specified in the given dt.  ASSUMES *left_cut, *right_cut = 0.
int hyp_OuterCutsOfSingleVertex(//INPUTS:
				vertex v,  hypergraph * g, inverse_hypergraph * h, 
				hyp_LCAinfo ** lca_info,
#ifdef OUTER_CUT_INFO
				dt_node * dt,
#endif

				//OUTPUTS:
				int * left_cut, int * right_cut)
{
  int i,j;
  hyp_LCAinfo * current_info;
  hyp_LCAinfo * right_child;
  hyp_LCAinfo * left_child;
  int * pi;
  int hyp;

#ifdef OVERFLOW_TEST
  int left_cut_update = 0; 
  int right_cut_update = 0; 
  
#endif



#ifdef OUTER_CUT_INFO
  printf("\nComputing Outer Cuts of Vertex %i under partition:\n",v);
  pi = hyp_DerivePermutation(dt,g);

  for(i=0; i<g->size; i++)
    if(pi[i] == v)
      printf(" ||| %i ||| ",v);
    else
      printf(" %i ",pi[i]);
  printf("\n");

  printf("   NumberOfHyperedgesCoveringVertex %i: %i\n",v,NumberOfHyperedgesCoveringVertex(v,h));

  for(i=0; i<NumberOfHyperedgesCoveringVertex(v,h); i++)
    {
      printf("      ");
      hyp = h->hyperedge_array[h->vertex_array[v] + i];
      for(j=g->hyperedge_array[hyp]; j<g->hyperedge_array[hyp+1]; j++)
	printf(" %i ",g->vertex_array[j]);
      printf("\n");
    }

#endif

  //For each hyperedge in which v is involved
  for(i=0; i<NumberOfHyperedgesCoveringVertex(v,h); i++)
    {

#ifdef OVERFLOW_TEST
      left_cut_update = 0;
      right_cut_update = 0;
#endif

      //find the lca info for the i-th hyperedge in which v is involved
      current_info = lca_info[h->hyperedge_array[h->vertex_array[v] + i]];
      
#ifdef OVERFLOW_TEST
      if(current_info == NULL) printf("The info is already NULL\n");
#endif

      //Find out if this hyperedge is in the left cut
      while(current_info != NULL)
	{
	  if(current_info->corr_dt->v == v)
	  {
	    (*left_cut) += 1;  //UNWEIGHTED
	    
#ifdef OVERFLOW_TEST
	    left_cut_update = 1;
#endif

	    current_info = NULL;
	  }
	  else if(current_info->corr_dt->orientation == -1)
	    current_info = NULL;
	  else
	    current_info = current_info->child[1 - current_info->corr_dt->orientation];   //LEFT CUT
	}

      //find the lca info for the i-th hyperedge in which v is involved
      current_info = lca_info[h->hyperedge_array[h->vertex_array[v] + i]];

      //Find out if this hyperedge is in the right cut
      while(current_info != NULL)
	{
	  if(current_info->corr_dt->v == v)
	    {
	
#ifdef OVERFLOW_TEST
	    right_cut_update = 1;
#endif
	    (*right_cut) += 1;  //UNWEIGHTED
	      current_info = NULL;
	    }
	  else if(current_info->corr_dt->orientation == -1)
	    current_info = NULL;
	  else
	    current_info = current_info->child[current_info->corr_dt->orientation];   //RIGHT CUT
	}
    

#ifdef OVERFLOW_TEST
      if ((!left_cut_update) && (!right_cut_update)) 
	{
	  
	  
	  printf("Edge (%i,%i) not in either cut!\n",
		 g->vertex_array[g->hyperedge_array[h->hyperedge_array[h->vertex_array[v] + i]]],
		 g->vertex_array[g->hyperedge_array[h->hyperedge_array[h->vertex_array[v] + i]]+1]);
	  
	  
	  //find the lca info for the i-th hyperedge in which v is involved
	  current_info = lca_info[h->hyperedge_array[h->vertex_array[v] + i]];
	  
	  printf("current info: v:%i, orient:%i\n",current_info->corr_dt->v,
		 current_info->corr_dt->orientation);
	  
	  left_child = current_info->child[0];
	  right_child = current_info->child[1];
	  
	  if(left_child !=NULL)
	    printf("left child: v:%i, orient:%i\n",left_child->corr_dt->v,
		   left_child->corr_dt->orientation);
	  else
	    printf("left child null\n");

	  if(right_child !=NULL)
	    printf("right child: v:%i, orient:%i\n",right_child->corr_dt->v,
		   right_child->corr_dt->orientation);
	  else
	    printf("right child null\n");

	  
	}
	  
	  

  
#endif
  
    }

  return 0;
}












