#include <time.h>
#include <stdio.h>
#include <math.h>
#include "ot.h"
#include "hyperorient.h"




//BINARY TO TAKE A RANDOM GRAPH with an input (n) number of nodes (2^n) and  
//orient it using HMETIS
int OrientRandomGraphUsingBalancedTree(int argc, char **argv)
{
  int cost,cost2;
  hypergraph * g;
  dt_node * dt;
  dt_node * dt2;
  

  if(argc < 2)
    {
      printf("Parameters:\n$1: depth of tree = log(nodes in graph)\n$2:LocalImprove UB Factor\n$3:Local Improve Trials\n$4:output filename\n");
    }

  g = NewRandomHypergraph(power2(atoi(argv[1])));
  
  dt = NewBalancedBinaryTree(g,atoi(argv[1]));

  cost = hyp_OrientAndEvaluateOnTheFly(g, dt);

  printf("Cost: %i\n",cost);

  dt2 = hyp_LocalImproveIterate(g,dt,atoi(argv[2]),atoi(argv[3]),&cost2);

  //  hyp_OutputHypergraph(g);

  hyp_OutputPermutationToFile(argv[4], dt2, g);
  
  /*  hyp_ComputeOptimalExhaustively(dt, g); */


  FreeDt(dt2);
  FreeHypergraph(g);
}


int CompareBestAndWorst(int argc, char **argv)
{

  FILE * fp;
  dt_node * dt;
  or_tree * best_or,*worst_or,*zero_or,*random_or;
  int best_cost,worst_cost,zero_cost,random_cost;
  hypergraph g;
  int cost,i,j;
  
  if(argc < 3)
    {  
      printf("Parameters:\n$1:Hypergraph File (HMETIS format)\n$2:UB Factor\n");
      return 0;
    }




  fp = fopen(argv[1],"r");

  if(fp == NULL)
    {
      printf("File %s Not Found\n",argv[1]);
      return 0;
      }
  
  printf("Zero Cost, Random Cost, Worst, Best\n");
  

      dt = create_decomposition_tree(fp, // the given hyper-graph
				     atoi(argv[2]), // relative imbalance factor
				     &g); // Hypergraph structure  
      
      
      best_or = hyp_OrientOnTheFly(&g, dt);
      OrientDtFromOrTree(dt, best_or);
      best_cost = hyp_CostOfOrientation(dt, &g);

      worst_or = hyp_OrientWorstOnTheFly(&g, dt);
      OrientDtFromOrTree(dt, worst_or);
      worst_cost = hyp_CostOfOrientation(dt, &g);

      zero_or = CreateZeroOrientation(dt);
      OrientDtFromOrTree(dt, zero_or);
      zero_cost = hyp_CostOfOrientation(dt, &g);

      random_cost = 0;
      for(j=1; j<=10; j++)
	{
	  random_or = CreateRandomOrientation(dt);
	  OrientDtFromOrTree(dt, random_or);
	  random_cost += hyp_CostOfOrientation(dt, &g);
	}
      random_cost = random_cost/10;

      printf("%i,%i,%i,%i\n",zero_cost,random_cost,worst_cost,best_cost);

      FreeDt(dt);
      
  
  
  FreeHypergraph(&g);


  fclose(fp);



}




int OrientLillisGraph(int argc, char **argv)
{
  

  FILE * fp;
  dt_node * dt;
  hypergraph g;
  int cost;

  
  if(argc < 3)
    {  
      printf("Parameters:\n$1:Hypergraph File (HMETIS format)\n$2:UB Factor\n$3:Output Filename for Layout\n");
      return 0;
    }




  fp = fopen(argv[1],"r");

  if(fp == NULL)
    {
      printf("File %s Not Found\n",argv[1]);
      return 0;
      }
  dt = create_decomposition_tree(fp, // the given hyper-graph
				 atoi(argv[2]), // relative imbalance factor
				 &g); // Hypergraph structure  
  
  cost = hyp_OrientAndEvaluateOnTheFly(&g, dt);


  printf("Cost: %i\n", cost);


  

  hyp_OutputPermutationToFile(argv[3], dt, &g);


  FreeHypergraph(&g);
  FreeDt(dt);

  fclose(fp);
}



int OrientAndImproveLillisGraph(int argc, char ** argv)
{
  int trials;
  int i,j,time_to_create_tree,time_to_orient,s,t,min;
  int ot_size;
  dt_node * dtmin = NULL,*dt2;
  FILE * fp;
  dt_node * dt;
  hypergraph * g;
  int cost;

  if(argc < 8)
    {  
      printf("Parameters:\n$1:Hypergraph File (HMETIS format)\n$2:UB Factor\n$3:Trials\n");
      printf("$4:Local Improvement UB Factor\n$5:Local Improvement trials\n");
      printf("$6:Upper Bound on OT size\n$7:(1)=Keep Hyperedge Projections\n$8:Output Filename for Layout\n");
      return 0;
    }

  g = CreateHypergraphFromFile(argv[1]);


#ifdef GRAPH_TESTING
  
  TestHypergraphForMultiedgesAndSelfLoops(g);


#endif



  if(atoi(argv[7]) == 1)
    SEPARATE_FLAG = 1;
  else
    SEPARATE_FLAG = 0;


  trials = atoi(argv[3]);
  printf("# HMETIS Decomposition and optimal Orientation:\n");
  printf("# trial, ot_size, score, time for dec tree, time for orienting\n") ;
  
  for(i = 1; i<=trials; i++)
    {
      
      s = time(NULL);
            
      dt = CreateDtFromHypergraph(atoi(argv[2]), g);

      t = time(NULL);
      time_to_create_tree = t - s;      


      ot_size = SizeOfOt(dt,0);
     
      
      if((ot_size > atoi(argv[6])) && (atoi(argv[6]) != -1))
	{
	  printf("OT Larger than upper threshold. Exiting\n%i\n",-1);
	  return -1;
	}

      printf("%i, %i, ",i,ot_size);

      s = time(NULL);
      cost = hyp_OrientAndEvaluateOnTheFly(g, dt);
      t = time(NULL);
      time_to_orient = t-s;      
      
      printf("%i, %i, %i\n",cost,time_to_create_tree,time_to_orient);

      if ((min > cost) || (i==1)) 
	{
	  min = cost;
	  if (dtmin != NULL) FreeDt(dtmin);
	  dtmin = dt;
	}
      else
	FreeDt(dt); 
      
    }
  
  printf("#Minimum: %i\n\n",min);
  printf("#Local Improvement:\n");
  
  
  dt2 = hyp_LocalImproveIterate(g,dtmin,atoi(argv[4]),atoi(argv[5]),&cost);
  
  
  
#ifdef OUTPUT_PERMUTATION    
  hyp_OutputPermutationToFile(argv[8],dt2,g);
#endif
  
  printf("%i\n",cost);
  
  FreeHypergraph(g);
  
}


int TestHypergraphForMultiedgesAndSelfLoops(hypergraph * g)
{


  int i,u,v,w;
  adjacency * ptr,*oldptr;
  adjacency ** a = NewArray(adjacency*, g->size);

  for(i=0; i<g->size; i++)
    a[i] = NULL;

  printf("Testing Hypergraph\n");

  for(i=0; i<g->hyperedges; i++)
    {
      u = g->vertex_array[g->hyperedge_array[i]];
      v = g->vertex_array[g->hyperedge_array[i]+1];

      //      printf("%i,%i\n",u,v);
      
      if(u == v)
	printf("SELF-LOOP: %i\n",u);

      if(u > v) { w=u; u=v; v=w; }
      
      
      if(a[u] == NULL) 
	{
	  a[u] = New(adjacency);
	  a[u]->node = v;
	  a[u]->next = NULL;
	}
      else
	{
	  //search a[u] for v
	  for(ptr = a[u]; ptr != NULL; ptr = ptr->next)
	    {
	      if(ptr->node == v)
		printf("MULTIEDGE: %i, %i\n",u,v);
	      oldptr = ptr;
	    }
	  
	  //insert a[u] at end
	  oldptr->next = New(adjacency);
	  oldptr->next->node = v;
	  oldptr->next->next = NULL;
	  
	}
      
      
    }
  
}









